import os

import click

import komodo_cli.printing as printing
from komodo_cli.backends.backend_factory import BackendFactory
from komodo_cli.utils import APIClient, handle_errors


@click.command("create")
@click.option(
    "--name",
    "-n",
    type=str,
    default=None,
    help="Name of the backend to create",
)
@click.option(
    "--backend-type",
    "-t",
    type=click.Choice(list(BackendFactory.backend_types.keys())),
    help="Type of backend to create.",
    required=True,
)
@click.pass_context
@handle_errors
def cmd_create(
    ctx: click.Context,
    name: str,
    backend_type: str,
):
    api_client: APIClient = ctx.obj["api_client"]

    backend_cls = BackendFactory.backend_types[backend_type]

    config_params = backend_cls.config_params
    config = {}
    for param in config_params:
        while True:
            value = click.prompt(param.name, default=None, type=param.dtype)

            if param.read_from_file:
                value = os.path.expanduser(value)
                if not os.path.isfile(value):
                    printing.error(f"{value} is not a valid file")
                    continue

                with open(value, "r") as f:
                    value = f.read()

            break

        config[param.name] = value

    api_client.create_backend(
        name,
        backend_type,
        config,
    )

    printing.success(f"Backend '{name}' succesfully created")



import click
from loguru import logger
from tabulate import tabulate

import komodo_cli.printing as printing
from komodo_cli.utils import (APIClient, handle_errors,
                              update_context_with_api_client)


@click.command("apply-promo-code")
@click.argument(
    "code",
    type=str,
)
@click.pass_context
@handle_errors
def cmd_apply_promo_code(ctx: click.Context, code: str):
    """List Komodo jobs."""
    logger.info("Listing jobs")

    update_context_with_api_client(ctx)
    api_client: APIClient = ctx.obj["api_client"]

    api_client.apply_promo_code(code)
    printing.success(f"Promo code {code} successfully applied")



import click
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.backends.backend import Backend
from komodo_cli.backends.backend_factory import BackendFactory
from komodo_cli.backends.local_backend import (JobNotFoundException,
                                               JobNotInDesiredStateException)
from komodo_cli.utils import APIClient, handle_errors


@click.command("terminate")
@click.argument(
    "name",
    type=str,
)
@click.option(
    "--force",
    "-f",
    type=bool,
    is_flag=True,
    help="Delete the machine even if it cannot be reached. This can result in a machine continuing to run on your infrastructure, but no longer being tracked by Komodo. Use with caution.",
)
@click.pass_context
@handle_errors
def cmd_terminate(ctx: click.Context, name: str, force: bool):
    logger.info(f"Machine Name: {name}")

    api_client: APIClient = ctx.obj["api_client"]

    machine = api_client.get_machine(name)
    backend_schema = api_client.get_backend(machine.backend_name)
    backend: Backend = BackendFactory.get_backend(
        backend_schema,
        api_client,
    )

    printing.warning(f"Terminating machine {name}", bold=True)
    api_client.delete_machine(machine.name, force)
    # only the local backend actually does anything here
    try:
        backend.cancel(machine.backend_job_id)
    except JobNotFoundException as e:
        if not force:
            raise e



import os
import sys

import click
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.commands.compute.utils import (get_compute_dir, get_vars,
                                               get_vars_filepath, run_cmd,
                                               update_vars)
from komodo_cli.utils import handle_errors


@click.command("add-aws-instance")
@click.option(
    "--name",
    "-n",
    type=str,
    required=True,
    default=None,
    help="Name of the compute environment to add.",
)
@click.option(
    "--instance-type",
    "-t",
    type=str,
    required=True,
    default=None,
    help="Instance type to add.",
)
@click.option(
    "--desired-vcpus",
    "-d",
    type=int,
    required=True,
    default=0,
    help="Desired number of vcpus for this compute environment.",
)
@click.option(
    "--min-vcpus",
    "-l",
    type=int,
    required=True,
    default=0,
    help="Mininum number of vcpus for this compute environment.",
)
@click.option(
    "--max-vcpus",
    "-h",
    type=int,
    required=True,
    default=10000,
    help="Maximum number of vcpus for this compute environment.",
)
@click.option(
    "--volume-size",
    "-v",
    type=int,
    required=True,
    default=100,
    help="Size of the volume in GB.",
)
@click.option(
    "--manual_confirm",
    "-m",
    type=bool,
    is_flag=True,
    default=False,
    required=False,
    help="Manually confirm terraform plan.",
)
@click.pass_context
@handle_errors
def cmd_add_compute_environment(
    ctx: click.Context,
    name: str,
    instance_type: str,
    desired_vcpus: int,
    min_vcpus: int,
    max_vcpus: int,
    volume_size: int,
    manual_confirm: bool,
):
    """
    Add a compute environment to AWS Batch.
    """
    compute_dir = get_compute_dir()
    if not os.path.exists(compute_dir):
        printing.error(f"No compute configuration found.", bold=True)
        sys.exit(1)

    try:
        tf_vars = get_vars()
    except Exception as e:
        logger.error(e)
        printing.error(f"Error loading tf vars", bold=True)
        sys.exit(1)

    if "aws_batch_compute_environments" not in tf_vars:
        tf_vars["aws_batch_compute_environments"] = {}

    tf_vars["aws_batch_compute_environments"][name] = {
        "instance_type": instance_type,
        "desired_vcpus": desired_vcpus,
        "min_vcpus": min_vcpus,
        "max_vcpus": max_vcpus,
        "volume_size": volume_size,
    }

    update_vars(tf_vars)

    cmd = [
        "terraform",
        f"-chdir={compute_dir}",
        "apply",
        f"-var-file={get_vars_filepath()}",
    ]

    if not manual_confirm:
        cmd.append("-auto-approve")

    exit_code, _, stderr = run_cmd(" ".join(cmd))
    if exit_code != 0:
        logger.error(stderr)
        printing.error(
            f"Error adding compute environment {name}",
            bold=True,
        )
        del tf_vars["aws_batch_compute_environments"][name]
        update_vars(tf_vars)
        sys.exit(1)

    printing.success(
        f"Successfully added compute environment {name}",
        bold=True,
    )



import os
import subprocess
from typing import Dict, Optional, Tuple
from uuid import getnode

import docker
from loguru import logger

from komodo_cli.backends.backend import Backend
from komodo_cli.image_builder.image_builder import ImageBuilder
from komodo_cli.types import (LOCAL_STATUS_MAPPING, ClientException, Job,
                              JobNotFoundException,
                              JobNotInDesiredStateException, JobStatus,
                              Machine, TooManyNodes)
from komodo_cli.utils import APIClient


class LocalBackend(Backend):
    config_params = ()
    resource_config_params = ()

    def __init__(self, name: str, api_client: APIClient, config: Dict, resources: Dict):
        super().__init__(name, api_client, config, resources)
        self.client = docker.from_env()

    def _docker_run(
        self,
        image,
        command,
        env,
        mounts,
        workdir,
        auto_remove=False,
    ):
        env["NODE_INDEX"] = 0
        env["NNODES"] = 1
        env["NODE_0_IP_ADDRESS"] = "localhost"
        container = self.client.containers.run(
            image=image,
            command=[
                "sh",
                "-c",
                " ".join(
                    f'"{a}"' if a not in ["&", "&&", "|", "||"] else a for a in command
                ),
            ],
            auto_remove=auto_remove,
            # cap_add=
            # cpu_count=
            # cpu_percent=
            detach=True,
            # devices=[],
            # https://stackoverflow.com/questions/71429711/how-to-run-a-docker-container-with-specific-gpus-using-docker-sdk-for-python
            # device_requests=[docker.types.DeviceRequest(device_ids=["0,2"], capabilities=[['gpu']])],
            environment=env,
            # mem_limit= #100000b, 1000k, 128m, 1g
            mounts=[
                docker.types.Mount(
                    target=m["target"],
                    source=m["source"] if m["source"] != "." else os.getcwd(),
                    type=m["type"],
                    read_only=False,
                )
                for m in mounts
            ],
            privileged=False,  # Does this need to be True to access devices?
            # runtime="nvidia",
            working_dir=workdir,
        )

        return container

    def run(
        self,
        command: Tuple[str],
        num_nodes: int,
        resource_name: str,
        image: str,
        env: Dict[str, str],
        mounts: list,
        workdir: str,
    ) -> Job:
        if num_nodes > 1:
            raise TooManyNodes(1)

        job = self.api_client.create_job(
            self.name,
            command,
            num_nodes,
            image,
            env,
            mounts,
            workdir,
            resource_name,
        )

        komodo_job_id = job.id

        try:
            container = self._docker_run(
                image,
                command,
                env,
                mounts,
                workdir,
            )
            backend_job_id = f"{getnode()}:{container.id[:12]}"
        except Exception as e:
            logger.error(f"Error starting container: {str(e)}")
            self.api_client.update_job(
                komodo_job_id,
                LOCAL_STATUS_MAPPING["dead"],  # TODO: need a failed job status?
            )
            raise e

        self.api_client.update_job(
            komodo_job_id,
            LOCAL_STATUS_MAPPING[container.status],
            backend_job_id,
        )

        return job

    def create_machine(
        self,
        machine_name: str,
        resource_name: dict,
        image: str,
        env: Dict[str, str],
        mounts: list,
        workdir: str,
    ) -> Machine:
        machine = self.api_client.create_machine(
            machine_name,
            self.name,
            image,
            env,
            mounts,
            workdir,
            resource_name,
        )

        try:
            container = self._docker_run(
                image,
                ["tail", "-f", "/dev/null"],
                env,
                mounts,
                workdir,
                True,
            )
            backend_job_id = f"{getnode()}:{container.id[:12]}"
        except Exception as e:
            logger.error(f"Error starting container: {str(e)}")
            self.api_client.update_machine(
                machine_name,
                LOCAL_STATUS_MAPPING["dead"],  # TODO: need failed job status?
            )
            raise e

        self.api_client.update_machine(
            machine_name,
            LOCAL_STATUS_MAPPING[container.status],
            backend_job_id,
        )

        return machine

    def logs(self, job_id: str, node_index: int, watch: bool):
        try:
            if ":" in job_id:
                mac_address, container_id = job_id.split(":")
            else:
                mac_address = None
                container_id = job_id

            if mac_address and mac_address != str(getnode()):
                raise ClientException(
                    f"Cannot get logs for job because it was not started on this machine"
                )
            container = self.client.containers.get(container_id)
        except docker.errors.NotFound:
            raise JobNotFoundException
        return container.logs(follow=watch, stream=True)

    def shell(self, job_id: str, node_index: int):
        if ":" in job_id:
            mac_address, container_id = job_id.split(":")
        else:
            mac_address = None
            container_id = job_id

        if mac_address and mac_address != str(getnode()):
            raise ClientException(
                f"Cannot shell into job because it was not started on this computer"
            )

        # /bin/sh links to the installed shell, whatever it is
        subprocess.call(["docker", "exec", "-it", container_id, "/bin/sh"])

    def vscode(self, job_id: str, node_index: int):
        if ":" in job_id:
            mac_address, container_id = job_id.split(":")
        else:
            mac_address = None
            container_id = job_id

        if mac_address and mac_address != str(getnode()):
            raise ClientException(
                f"Cannot shell into job because it was not started on this computer"
            )

        container = self.client.containers.get(container_id)
        container_info = container.attrs
        workdir = container_info["Config"]["WorkingDir"]
        subprocess.call(
            [
                "bash",
                "-c",
                f'code --folder-uri vscode-remote://attached-container+$(printf "{container_id}" | xxd -p){workdir}',
            ]
        )

    def cancel(self, job_id: str):
        try:
            if ":" in job_id:
                mac_address, container_id = job_id.split(":")
            else:
                mac_address = None
                container_id = job_id

            if mac_address and mac_address != str(getnode()):
                raise ClientException(
                    f"Cannot cancel job because it was not started on this machine"
                )
            container = self.client.containers.get(container_id)
        except docker.errors.NotFound:
            raise JobNotFoundException
        except docker.errors.NullResource:
            raise JobNotFoundException
        container.kill()

    def delete(self, job_id: str):
        try:
            if ":" in job_id:
                mac_address, container_id = job_id.split(":")
            else:
                mac_address = None
                container_id = job_id

            if mac_address and mac_address != str(getnode()):
                raise ClientException(
                    f"Cannot cancel job because it was not started on this machine"
                )
            container = self.client.containers.get(container_id)
        except docker.errors.NotFound:
            raise JobNotFoundException
        except docker.errors.NullResource:
            raise JobNotFoundException
        container.remove(force=True)

    @staticmethod
    def supports_shell() -> bool:
        return True

    def prepare_image(
        self,
        base_image: str,
        project_dir: str,
        workspace: Optional[str],
        workdir: Optional[str],
    ) -> str:
        builder = ImageBuilder(base_image, project_dir)

        if workspace:
            if not workdir:
                raise ClientException(
                    "Workspace was provided but no workdir was provided"
                )
            builder.add_overlay(workspace, workdir)

        if workdir:
            builder.set_workdir(workdir)

        image = builder.build_image()
        return image.tags[0]



import click

from komodo_cli.commands.backend.cmd_add_resource import cmd_add_resource
from komodo_cli.commands.backend.cmd_create import cmd_create
from komodo_cli.commands.backend.cmd_delete import cmd_delete
from komodo_cli.commands.backend.cmd_delete_resource import cmd_delete_resource
from komodo_cli.commands.backend.cmd_list import cmd_list
from komodo_cli.utils import update_context_with_api_client


@click.group()
@click.pass_context
def backend(ctx: click.Context):
    update_context_with_api_client(ctx)


backend.add_command(cmd_create)
backend.add_command(cmd_list)
backend.add_command(cmd_delete)
backend.add_command(cmd_add_resource)
backend.add_command(cmd_delete_resource)



import click

from komodo_cli.commands.init.cmd_project import cmd_project
from komodo_cli.commands.init.cmd_task import cmd_task
from komodo_cli.utils import update_context_with_api_client


@click.group()
@click.pass_context
def init(ctx: click.Context):
    update_context_with_api_client(ctx)


init.add_command(cmd_project)
init.add_command(cmd_task)



import click

import komodo_cli.printing as printing
from komodo_cli.backends.backend_factory import BackendFactory
from komodo_cli.utils import APIClient, handle_errors


@click.command("add-resource")
@click.option(
    "--backend",
    "-b",
    type=str,
    required=True,
    help="Name of the backend to create the resource under",
)
@click.option(
    "--name",
    "-n",
    type=str,
    required=True,
    help="Name of the resource to create",
)
@click.pass_context
@handle_errors
def cmd_add_resource(
    ctx: click.Context,
    backend: str,
    name: str,
):
    api_client: APIClient = ctx.obj["api_client"]

    backend_name = backend
    backend = api_client.get_backend(backend_name)
    backend_cls = BackendFactory.backend_types[backend.type]

    resource_config_params = backend_cls.resource_config_params
    resource_config = {}
    for param in resource_config_params:
        if param.required:
            text = param.name
            default = None
            show_default = True
        else:
            text = f"{param.name} (optional)"
            default = ""
            show_default = False

        value = click.prompt(
            text, default=default, show_default=show_default, type=param.dtype
        )
        if value == "":
            value = None
        resource_config[param.name] = value

    api_client.create_backend_resource(
        backend_name,
        name,
        resource_config,
    )

    printing.success(
        f"Resource '{name}' succesfully created for backend '{backend_name}'"
    )



import os
import stat
import subprocess
import tempfile
from typing import Dict, Optional, Tuple

import boto3
import botocore
from loguru import logger

from komodo_cli.backends.backend import Backend, ConfigParam
from komodo_cli.image_builder.image_builder import ImageBuilder
from komodo_cli.types import ClientException, Job, Machine
from komodo_cli.utils import APIClient


class AWSBackend(Backend):
    config_params = (
        ConfigParam("access_key_id", str, "AWS access key ID", True),
        ConfigParam("secret_access_key", str, "AWS secret access key", True),
        ConfigParam("region", str, "AWS region", True),
        ConfigParam(
            "ssh-key",
            str,
            "Path to the ssh key needed to allow ssh access to the EC2 instances",
            True,
            True,
        ),
        ConfigParam(
            "docker_repo",
            str,
            "The Docker registry (ideally an ECR repo) used to store job images",
            True,
        ),
        ConfigParam(
            "bastion-ip-address",
            str,
            "The public IP address of the bastion EC2 instance.",
            True,
        ),
    )
    resource_config_params = (
        ConfigParam("instance_type", str, "instance type", True),
        ConfigParam("aws_batch_queue", str, "name of the AWS Batch queue", True),
    )

    def __init__(self, name: str, api_client: APIClient, config: Dict, resources: Dict):
        super().__init__(name, api_client, config, resources)

        session = boto3.Session(
            region_name=config["region"],
        )
        self._batch_client = session.client("batch", region_name=config["region"])
        self._ec2_client = session.client("ec2", region_name=config["region"])
        self._logs_client = session.client("logs", region_name=config["region"])
        self._ecs_client = session.client("ecs", region_name=config["region"])

    def run(
        self,
        command: Tuple[str],
        num_nodes: int,
        resource_name: str,
        image: str,
        env: Dict[str, str],
        mounts: list,
        workdir: str,
    ) -> Job:
        job = self.api_client.create_job(
            self.name,
            command,
            num_nodes,
            image,
            env,
            mounts,
            workdir,
            resource_name,
        )

        return job

    def create_machine(
        self,
        machine_name: str,
        resource_name: str,
        image: str,
        env: Dict[str, str],
        mounts: list,
        workdir: str,
    ) -> Machine:
        machine = self.api_client.create_machine(
            machine_name,
            self.name,
            image,
            env,
            mounts,
            workdir,
            resource_name,
        )
        return machine

    def _get_job(self, job_id: str, node_index: int) -> dict:
        job_id = f"{job_id}#{node_index}"
        try:
            jobs = self._batch_client.describe_jobs(jobs=[job_id])["jobs"]
            if len(jobs) == 0:
                raise ClientException(f"AWS Batch job {job_id} not found")

            assert len(jobs) == 1
            job = jobs[0]

            return job
        except botocore.exceptions.ClientError as err:
            logger.error(str(err))
            raise err

    def logs(self, job_id: str, node_index: int, watch: bool):
        job = self._get_job(job_id, node_index)

        if job.get("status", None) == "RUNNING":
            log_stream_name = job["container"]["logStreamName"]
        else:
            attempts = job["attempts"]
            if len(attempts) == 0:
                return

            attempt = attempts[-1]
            container = attempt["container"]
            log_stream_name = container["logStreamName"]

        next_token = None

        while True:
            args = {}
            if next_token is not None:
                args["nextToken"] = next_token

            try:
                response = self._logs_client.get_log_events(
                    logGroupName="/aws/batch/job",
                    logStreamName=log_stream_name,
                    limit=10000,
                    startFromHead=True,
                    **args,
                )
            except self._logs_client.exceptions.ResourceNotFoundException:
                return
            if response["nextForwardToken"] == next_token:
                # we've reached the end of the available logs

                # if watch is False, then we end here
                if not watch:
                    break

                # if the job is running, we wait for more logs
                job = self._get_job(job_id, node_index)
                if job["status"] == "RUNNING":
                    continue

                # if the job is not running, then we end here
                break

            next_token = response["nextForwardToken"]

            for event in response["events"]:
                yield event["message"] + "\n"

    def shell(self, job_id: str, node_index: int):
        job = self._get_job(job_id, node_index)
        status = job["status"]
        if status != "RUNNING":
            raise ClientException("Job is not running, cannot shell into it")

        task_arn = job["container"]["taskArn"]
        job_queue_arn = job["jobQueue"]
        job_queue = self._batch_client.describe_job_queues(jobQueues=[job_queue_arn])[
            "jobQueues"
        ][0]

        # these are all the possible compute environments that the job could've been scheduled on
        possible_compute_environment_arns = [
            e["computeEnvironment"] for e in job_queue["computeEnvironmentOrder"]
        ]
        possible_compute_environments = (
            self._batch_client.describe_compute_environments(
                computeEnvironments=possible_compute_environment_arns,
            )["computeEnvironments"]
        )

        # these are all the possible ECS clusters (corresponding to AWS Batch compute environments) that
        # the job could be running in
        possible_ecs_cluster_arns = [
            e["ecsClusterArn"] for e in possible_compute_environments
        ]

        job_cluster_arn = None
        task = None
        for cluster_arn in possible_ecs_cluster_arns:
            tasks = self._ecs_client.describe_tasks(
                cluster=cluster_arn,
                tasks=[task_arn],
            )["tasks"]

            if len(tasks) == 0:
                continue

            job_cluster_arn = cluster_arn
            task = tasks[0]
            break

        if job_cluster_arn is None:
            raise ClientException(
                "Could not find the corresponding ECS cluster for the job"
            )

        container_arn = task["containerInstanceArn"]
        container_instance = self._ecs_client.describe_container_instances(
            containerInstances=[container_arn],
            cluster=job_cluster_arn,
        )["containerInstances"][0]

        ec2_instance_id = container_instance["ec2InstanceId"]

        instance = self._ec2_client.describe_instances(
            InstanceIds=[ec2_instance_id],
        )[
            "Reservations"
        ][0]["Instances"][0]

        private_ip = instance.get("PrivateDnsName", None)
        if private_ip is None:
            raise ClientException("Cannot find IP address for instance")

        key = tempfile.mktemp()
        with open(key, "w") as f:
            f.write(self.config["ssh-key"])
        os.chmod(key, stat.S_IRUSR)

        p = subprocess.Popen(
            [
                "ssh",
                "-vvv",
                "-i",
                key,
                "-o",
                "StrictHostKeyChecking=no",
                "-o",
                "IdentitiesOnly=yes",
                f"ec2-user@{private_ip}",
                "-o",
                f"proxycommand ssh -W %h:%p -i {key} -o IdentitiesOnly=yes ec2-user@{self.config['bastion-ip-address']}",
                "docker",
                "ps",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate()
        rc = p.returncode

        if rc != 0:
            raise ClientException(f"Got error: \n{stderr.decode('utf-8')}")

        stdout = stdout.decode("utf-8")
        lines = stdout.splitlines()
        if lines[0].split()[0] != "CONTAINER":
            raise ClientException("Cannot find docker container")

        num_amazon_containers = 0
        non_amazon_containers = []
        for line in lines[1:]:
            if not line:
                continue
            parts = line.split(" ")
            parts = [p for p in parts if p]

            container_id = parts[0]
            image = parts[1]
            if image.startswith("amazon"):
                num_amazon_containers += 1
            else:
                non_amazon_containers.append(container_id)

        if num_amazon_containers == 0 or len(non_amazon_containers) != 1:
            raise ClientException("Cannot find docker container")

        container_id = non_amazon_containers[0]

        subprocess.call(
            [
                "ssh",
                "-t",
                "-i",
                key,
                "-o",
                "StrictHostKeyChecking=no",
                "-o",
                "IdentitiesOnly=yes",
                "-o",
                f"proxycommand ssh -W %h:%p -i {key} -o IdentitiesOnly=yes ec2-user@{self.config['bastion-ip-address']}",
                f"ec2-user@{private_ip}",
                f"docker exec -it {container_id} /bin/bash",
            ]
        )

        os.remove(key)

    def cancel(self, job_id: str):
        pass

    def delete(self, job_id: str):
        pass

    @staticmethod
    def supports_shell() -> bool:
        return True

    def prepare_image(
        self,
        base_image: str,
        project_dir: str,
        workspace: Optional[str],
        workdir: Optional[str],
    ) -> str:
        builder = ImageBuilder(base_image, project_dir)

        builder.add_aws_cli()
        builder.add_aws_efa()

        if workspace:
            if not workdir:
                raise ClientException(
                    "Workspace was provided but no workdir was provided"
                )
            builder.add_overlay(workspace, workdir)

        if workdir:
            builder.set_workdir(workdir)

        docker_repo = self.config["docker_repo"]
        image = builder.build_image(docker_repo)
        # log in to ECR
        if "dkr.ecr" in docker_repo:
            exit_status = os.system(
                f"aws ecr get-login-password --region {self.config['region']} | docker login --username AWS --password-stdin {docker_repo}"
            )
            if exit_status != 0:
                raise ClientException("Failed to log in with ECR")

        builder.push_image(image, docker_repo)
        return image.tags[0]



import sys

import click
from tabulate import tabulate

import komodo_cli.printing as printing
from komodo_cli.utils import APIClient, handle_errors


@click.command("list")
@click.pass_context
@handle_errors
def cmd_list(ctx: click.Context):
    api_client: APIClient = ctx.obj["api_client"]

    machines = api_client.list_machines()

    machines_to_print = [
        [
            machine.name,
            machine.status.value,
            machine.backend_name,
            machine.resource_name,
            machine.backend_job_id,
        ]
        for machine in machines
    ]

    printing.header(f"Found {len(machines)} Komodo machines\n", bold=True)
    printing.info(
        tabulate(
            machines_to_print,
            headers=["Name", "Status", "Backend", "Resource", "Backend Job ID"],
            tablefmt="simple_grid",
        ),
    )



import click

from komodo_cli.commands.api_key.cmd_create import cmd_create
from komodo_cli.commands.api_key.cmd_delete import cmd_delete
from komodo_cli.commands.api_key.cmd_list import cmd_list


@click.group(name="api-key")
@click.pass_context
def api_key(ctx: click.Context):
    pass


api_key.add_command(cmd_create)
api_key.add_command(cmd_list)
api_key.add_command(cmd_delete)



import click

import komodo_cli.printing as printing
from komodo_cli.utils import (APIClient, handle_errors,
                              update_context_with_api_client)


@click.command("list")
@click.pass_context
@handle_errors
def cmd_list(ctx: click.Context):
    update_context_with_api_client(ctx)
    api_client: APIClient = ctx.obj["api_client"]

    tokens = api_client.get_all_tokens()

    printing.header(
        f"All of your API keys are below:\n",
    )
    printing.info("\n".join(tokens))



import os
import sys
from typing import Optional

import click
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.commands.compute.utils import (SUPPORTED_COMPUTE_SERVICES,
                                               get_compute_dir, get_vars,
                                               get_vars_filepath, run_cmd)
from komodo_cli.utils import handle_errors


@click.command("terminate")
@click.option(
    "--compute-type",
    "-t",
    type=str,
    default="aws",
    help="The type of compute infrastructure type to destroy.",
)
@click.option(
    "--name",
    "-n",
    type=str,
    required=True,
    help="The name of the compute infrastructure to destroy",
)
@click.option(
    "--manual-confirm",
    "-m",
    type=bool,
    is_flag=True,
    default=False,
    required=False,
    help="Manually confirm terraform plan.",
)
@click.pass_context
@handle_errors
def cmd_terminate(
    ctx: click.Context,
    compute_type: Optional[str],
    name: Optional[str],
    manual_confirm: Optional[bool],
):
    """Destroy Komodo compute"""
    if compute_type not in SUPPORTED_COMPUTE_SERVICES:
        printing.error(
            f"Unsupported compute type {compute_type}. Supported compute types: {SUPPORTED_COMPUTE_SERVICES}",
            bold=True,
        )
        sys.exit(1)

    compute_dir = get_compute_dir()

    if not os.path.exists(compute_dir):
        printing.error(f"No compute configuration found.", bold=True)
        sys.exit(1)

    try:
        tf_vars = get_vars()
    except Exception as e:
        logger.error(e)
        printing.error(f"Error loading tf vars", bold=True)
        sys.exit(1)

    if "eks_cluster_id" in tf_vars:
        if tf_vars["eks_cluster_id"][1:-1] != name:
            printing.error(
                f"Cluster ID {name} does not match the cluster ID in the configuration {tf_vars['eks_cluster_id']}",
                bold=True,
            )
            sys.exit(1)

    cmd = [
        "terraform",
        f"-chdir={compute_dir}",
        "destroy",
        f"-var-file={get_vars_filepath()}",
    ]

    if not manual_confirm:
        cmd.append("-auto-approve")

    exit_code, _, stderr = run_cmd(" ".join(cmd))
    if exit_code != 0:
        logger.error(stderr)
        printing.error("Error destroying compute", bold=True)
        sys.exit(1)

    os.remove(get_vars_filepath())
    printing.success("Compute destroyed successfully", bold=True)
    # TODO: delete backend from server via api call.



import click

from komodo_cli.commands.machine.cmd_create import cmd_create
from komodo_cli.commands.machine.cmd_list import cmd_list
from komodo_cli.commands.machine.cmd_shell import cmd_shell
from komodo_cli.commands.machine.cmd_terminate import cmd_terminate
from komodo_cli.commands.machine.cmd_vscode import cmd_vscode
from komodo_cli.utils import update_context_with_api_client


@click.group()
@click.pass_context
def machine(ctx: click.Context):
    update_context_with_api_client(ctx)


machine.add_command(cmd_create)
machine.add_command(cmd_list)
machine.add_command(cmd_terminate)
machine.add_command(cmd_shell)
machine.add_command(cmd_vscode)



import abc
from abc import ABC
from typing import Dict, NamedTuple, Optional, Tuple

from komodo_cli.types import Backend as BackendSchema
from komodo_cli.types import ClientException, Job, JobStatus, Machine
from komodo_cli.utils import APIClient


class ConfigParam(NamedTuple):
    name: str
    dtype: type
    description: str
    required: bool
    read_from_file: bool = False


class Backend(ABC):
    def __init__(self, name: str, api_client: APIClient, config: Dict, resources: Dict):
        self.name = name
        self.api_client: APIClient = api_client
        self.config = config
        self.resources = resources

    # The params expected in the config for the backend (eg. credentials)
    config_params: Tuple[ConfigParam, ...]
    # The keys expected in the configs of the resources added to the backend (eg. num gpus)
    resource_config_params: Tuple[ConfigParam, ...]

    @abc.abstractmethod
    def run(
        self,
        command: Tuple[str],
        num_nodes: int,
        resource_name: str,
        image: str,
        env: Dict[str, str],
        mounts: list,
        workdir: str,
    ) -> Job:
        pass

    @abc.abstractmethod
    def create_machine(
        self,
        machine_name: str,
        resource_name: dict,
        image: str,
        env: Dict[str, str],
        mounts: list,
        workdir: str,
    ) -> Machine:
        pass

    @abc.abstractmethod
    def logs(self, job_id: str, node_index: int, watch: bool):
        pass

    @abc.abstractmethod
    def shell(self, job_id: str, node_index: int):
        pass

    def vscode(self, job_id: str, node_index: int):
        raise ClientException(
            "Support for VSCode for this type of backend is not supported yet, but coming soon!"
        )

    @abc.abstractmethod
    def cancel(self, job_id: str):
        pass

    @abc.abstractmethod
    def delete(self, job_id: str):
        pass

    @abc.abstractstaticmethod
    def supports_shell() -> bool:
        pass

    @abc.abstractmethod
    def prepare_image(
        self,
        base_image: str,
        project_dir: str,
        workspace: Optional[str],
        workdir: Optional[str],
    ) -> str:
        pass



import click

import komodo_cli.printing as printing
from komodo_cli.utils import (APIClient, handle_errors,
                              update_context_with_api_client)


@click.command("delete")
@click.argument(
    "api-key",
    type=str,
)
@click.pass_context
@handle_errors
def cmd_delete(ctx: click.Context, api_key: str):
    update_context_with_api_client(ctx)
    api_client: APIClient = ctx.obj["api_client"]

    api_client.delete_token(api_key)
    printing.success(f"API Key {api_key} has been successfully deleted")



import click

from komodo_cli.commands.compute.cmd_add_compute_environment import \
    cmd_add_compute_environment
from komodo_cli.commands.compute.cmd_add_node_group import cmd_add_node_group
from komodo_cli.commands.compute.cmd_add_storage import cmd_add_storage
from komodo_cli.commands.compute.cmd_create import cmd_create
from komodo_cli.commands.compute.cmd_list import cmd_list
from komodo_cli.commands.compute.cmd_remove_node_group import \
    cmd_remove_node_group
from komodo_cli.commands.compute.cmd_remove_storage import cmd_remove_storage
from komodo_cli.commands.compute.cmd_terminate import cmd_terminate


@click.group()
@click.pass_context
def compute(ctx: click.Context):
    """Manage Komodo compute"""
    pass


compute.add_command(cmd_create)
compute.add_command(cmd_list)
compute.add_command(cmd_terminate)
# compute.add_command(cmd_add_storage)
# compute.add_command(cmd_remove_storage)
# compute.add_command(cmd_add_node_group)
# compute.add_command(cmd_remove_node_group)
compute.add_command(cmd_add_compute_environment)



import os
import sys
from typing import Optional

import click
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.commands.compute.utils import (get_compute_dir, get_vars,
                                               get_vars_filepath, run_cmd,
                                               update_vars)
from komodo_cli.utils import handle_errors


@click.command("remove-storage")
@click.option(
    "--storage-type",
    "-s",
    type=str,
    default="fsx",
    help="Storage type to add.",
)
@click.option(
    "--s3-bucket",
    type=str,
    required=True,
    default=None,
    help="Name of the S3 bucket backing the storage",
)
@click.option(
    "--manual_confirm",
    "-m",
    type=bool,
    is_flag=True,
    default=False,
    required=False,
    help="Manually confirm terraform plan.",
)
@click.pass_context
@handle_errors
def cmd_remove_storage(
    ctx: click.Context,
    storage_type: Optional[str],
    s3_bucket: Optional[str],
    manual_confirm: Optional[bool],
):
    compute_dir = get_compute_dir()

    if not os.path.exists(compute_dir):
        printing.error(f"No compute configuration found.", bold=True)
        sys.exit(1)

    try:
        tf_vars = get_vars()
    except Exception as e:
        logger.error(e)
        printing.error(
            f"Error loading {storage_type} for {s3_bucket}",
            bold=True,
        )
        sys.exit(1)

    if storage_type == "fsx":
        fsx_buckets = tf_vars.get("fsx_s3_buckets", [])
        for idx, curr in enumerate(fsx_buckets):
            if s3_bucket == curr:
                fsx_buckets.pop(idx)

        tf_vars["fsx_s3_buckets"] = fsx_buckets
        update_vars(tf_vars)

    cmd = [
        "terraform",
        f"-chdir={compute_dir}",
        "apply",
        f"-var-file={get_vars_filepath()}",
    ]

    if not manual_confirm:
        cmd.append("-auto-approve")

    exit_code, _, stderr = run_cmd(" ".join(cmd))

    if exit_code != 0:
        logger.error(stderr)
        printing.error(
            f"Error removing {storage_type} storage for {s3_bucket}",
            bold=True,
        )
        sys.exit(1)

    printing.success(
        f"Storage removed successfully: {storage_type}/{s3_bucket}",
        bold=True,
    )



import os

import click
import git
import jinja2

import komodo_cli.printing as printing
from komodo_cli.utils import APIClient

TASK_TEMPLATE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "../..", "templates"
)


@click.command("task")
@click.option(
    "--name",
    "-n",
    type=str,
    help="Name of task to create.",
)
@click.pass_context
def cmd_task(ctx: click.Context, name: str):
    git_repo = git.Repo(os.getcwd(), search_parent_directories=True)
    git_root = git_repo.git.rev_parse("--show-toplevel")

    task_config_filepath = os.path.join(git_root, ".komo", f"{name}.yaml")
    if os.path.isfile(task_config_filepath):
        printing.error(
            f"Project config already exists in this repository",
            bold=True,
        )
        exit(1)
    printing.info(
        f"Initializing Komo task config in {task_config_filepath}",
        bold=True,
    )

    api_client: APIClient = ctx.obj["api_client"]
    backend_schemas = api_client.list_backends()

    template_loader = jinja2.FileSystemLoader(
        searchpath=os.path.join(TASK_TEMPLATE_DIR)
    )
    template_env = jinja2.Environment(loader=template_loader)
    project_template = template_env.get_template("task.jinja")

    backends_str = " | ".join([b.name for b in backend_schemas])

    task_config_output = project_template.render(backends=backends_str)
    os.makedirs(os.path.dirname(task_config_filepath), exist_ok=True)
    with open(task_config_filepath, "w") as f:
        f.write(task_config_output)



import os
import time
from typing import Optional

import click
from loguru import logger
from vyper import Vyper

import komodo_cli.printing as printing
from komodo_cli.backends.backend import Backend
from komodo_cli.backends.backend_factory import BackendFactory
from komodo_cli.types import JobStatus
from komodo_cli.utils import (
    APIClient, config_args_to_dict, handle_errors,
    update_context_project_config_with_backend_override,
    update_context_with_project_config, validate_backend_config,
    validate_overlay_images_repository)


@click.command("create")
@click.argument(
    "name",
    type=str,
)
@click.option(
    "--backend",
    "-b",
    type=str,
    default=None,
    help="Name of the backend to use.",
)
@click.option(
    "--resource",
    "-r",
    type=str,
    default=None,
    help="Resource type to use for the job as defined in the config file. Ignored for local jobs.",
)
@click.option(
    "--config",
    "-c",
    type=str,
    default="",
    help=(
        "Override any parameters in the section of your project config corresponding to the specified backend."
        "Values must be provided as a comma-separated list of key=value pairs."
    ),
)
@click.option("--detach", "-d", is_flag=True)
@click.pass_context
@handle_errors
def cmd_create(
    ctx: click.Context,
    name: str,
    backend: Optional[str],
    resource: Optional[str],
    config: str,
    detach: bool,
):
    backend_name = backend
    update_context_with_project_config(ctx)
    api_client: APIClient = ctx.obj["api_client"]

    project_config: Vyper = ctx.obj["project_config"]
    if backend_name is None:
        backend_name = project_config.get("default_backend")
        if backend_name is None:
            printing.error("No backend was specified", bold=True)
            exit(1)

    backend_config_override = config_args_to_dict(config)
    update_context_project_config_with_backend_override(
        ctx, backend_name, backend_config_override
    )
    project_backend_config = project_config.get("configs")[backend_name]
    try:
        validate_backend_config(backend_name, project_backend_config)
    except Exception as e:
        printing.error(f"Error while validating backend config: {str(e)}", bold=True)
        exit(1)

    backend_schema = api_client.get_backend(backend_name)
    backend: Backend = BackendFactory.get_backend(
        backend_schema,
        api_client,
    )

    if resource is None:
        resource = project_backend_config.get("default_resource")
        if resource is None:
            # resource not needed for local backend
            if backend_schema.type != "local":
                printing.error("No resource was specified", bold=True)
                exit(1)

    logger.info(f"Using backend {backend_name}")
    logger.info(f"Resource: {resource}")

    if not backend.supports_shell():
        printing.error(
            f"Backend {backend_name} does not support machines",
            bold=True,
        )
        return

    workspace = project_backend_config.get("workspace", os.getcwd())
    workdir = project_backend_config.get("workdir", None)
    image = backend.prepare_image(
        project_backend_config["image"], ctx.obj["project_dir"], workspace, workdir
    )
    project_backend_config["image"] = image

    printing.header("Starting a machine", bold=True)
    machine = backend.create_machine(
        name,
        resource,
        project_backend_config["image"],
        project_backend_config["env"],
        project_backend_config["mounts"],
        project_backend_config["workdir"],
    )

    printing.success(
        f"Created a machine on backend {machine.backend_name} with name {machine.name}",
        bold=True,
    )

    if detach:
        return

    printing.success("Waiting for machine to start...", bold=True)

    while True:
        machine = api_client.get_machine(name)
        status = machine.status
        if status != JobStatus.PENDING:
            break
        time.sleep(1)

    backend.shell(machine.backend_job_id, 0)



from typing import Dict, Type

from komodo_cli.backends.aws_backend import AWSBackend
from komodo_cli.backends.backend import Backend
from komodo_cli.backends.kubernetes_backend import KubernetesBackend
from komodo_cli.backends.local_backend import LocalBackend
from komodo_cli.types import Backend as BackendSchema
from komodo_cli.utils import APIClient


class BackendFactory:
    backend_types: Dict[str, Type[Backend]] = {
        "local": LocalBackend,
        "kubernetes": KubernetesBackend,
        "aws": AWSBackend,
    }

    @classmethod
    def get_backend(
        cls, backend_schema: BackendSchema, api_client: APIClient
    ) -> Backend:
        resources_dict = {r.name: r.config for r in backend_schema.resources}
        backend = cls.backend_types[backend_schema.type](
            backend_schema.name,
            api_client,
            backend_schema.config,
            resources_dict,
        )
        return backend



import sys

import click
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.backends.backend import Backend
from komodo_cli.backends.backend_factory import BackendFactory
from komodo_cli.types import JobStatus
from komodo_cli.utils import APIClient, handle_errors


@click.command("vscode")
@click.argument(
    "name",
    type=str,
)
@click.pass_context
@handle_errors
def cmd_vscode(ctx: click.Context, name: str):
    logger.info(f"Machine: {name}")

    api_client: APIClient = ctx.obj["api_client"]

    machine = api_client.get_machine(name)

    if machine.status != JobStatus.RUNNING:
        printing.error(f"Machine {name} is not running", bold=True)
        exit(1)

    backend_schema = api_client.get_backend(machine.backend_name)
    backend: Backend = BackendFactory.get_backend(
        backend_schema,
        api_client,
    )

    printing.header(f"Opening a VSCode session into machine {name}", bold=True)
    backend.vscode(machine.backend_job_id, 0)



import click

import komodo_cli.printing as printing
from komodo_cli.backends.backend_factory import Backend, BackendFactory
from komodo_cli.types import JobStatus
from komodo_cli.utils import APIClient, handle_errors


@click.command("delete")
@click.option(
    "--backend",
    "-b",
    type=str,
    default=None,
    help="Name of the backend to delete",
)
@click.option(
    "--force",
    "-f",
    type=bool,
    is_flag=True,
    help="Delete the backend even if there are jobs/machines that are still running or unreacheable (those will be deleted, but not cancelled)",
)
@click.pass_context
@handle_errors
def cmd_delete(
    ctx: click.Context,
    backend: str,
    force: bool,
):
    backend_name = backend
    api_client: APIClient = ctx.obj["api_client"]

    printing.warning(
        "Deleting a backend will delete all jobs and machines associated with this backend. Are you sure you want to continue?",
        nl=False,
    )
    confirmed = click.confirm("")
    if not confirmed:
        return

    backend_schema = api_client.get_backend(backend_name)

    jobs = api_client.list_jobs()
    jobs = [j for j in jobs if j.backend_name == backend_name]
    machines = api_client.list_machines()
    machines = [m for m in machines if m.backend_name == backend_name]

    api_client.delete_backend(backend_name, force)

    # if we've gotten to this point, then all the jobs/machines associated with
    # this backend have been deleted
    # now let's cancel the running ones (only applies for local jobs, since server handles the rest)
    backend: Backend = BackendFactory.get_backend(backend_schema, api_client)
    for job in jobs:
        if job.status in [JobStatus.RUNNING, JobStatus.PENDING]:
            backend.cancel(job.backend_job_id)
    for machine in machines:
        if machine.status in [JobStatus.RUNNING, JobStatus.PENDING]:
            backend.cancel(machine.backend_job_id)

    printing.success(f"Backend '{backend_name}' succesfully deleted")



import os
import subprocess
from typing import Dict, Optional, Tuple

from kubernetes import client
from kubernetes import watch as k8swatch
from kubernetes.client.rest import ApiException
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.backends.backend import Backend, ClientException, ConfigParam
from komodo_cli.image_builder.image_builder import ImageBuilder
from komodo_cli.types import Job, Machine
from komodo_cli.utils import APIClient


class KubernetesBackend(Backend):
    config_params = (
        ConfigParam("host", str, "URL of the cluster", True),
        ConfigParam("bearer_token", str, "token to access the cluster", True),
        ConfigParam(
            "docker_repo",
            str,
            "The Docker registry used to store job images",
            True,
        ),
    )
    resource_config_params = (
        ConfigParam("gpus", int, "number of gpus", True),
        ConfigParam("cpus", float, "number of cpus", True),
        ConfigParam("memMB", int, "amount of RAM, in MB", True),
        ConfigParam("instance_type", str, "instance type", False),
    )

    def __init__(self, name: str, api_client: APIClient, config: Dict, resources: Dict):
        super().__init__(name, api_client, config, resources)

    def run(
        self,
        command: Tuple[str],
        num_nodes: int,
        resource_name: str,
        image: str,
        env: Dict[str, str],
        mounts: list,
        workdir: str,
    ) -> Job:
        job = self.api_client.create_job(
            self.name,
            command,
            num_nodes,
            image,
            env,
            mounts,
            workdir,
            resource_name,
        )

        return job

    def create_machine(
        self,
        machine_name: str,
        resource_name: str,
        image: str,
        env: Dict[str, str],
        mounts: list,
        workdir: str,
    ) -> Machine:
        machine = self.api_client.create_machine(
            machine_name,
            self.name,
            image,
            env,
            mounts,
            workdir,
            resource_name,
        )
        return machine

    def logs(self, job_id: str, node_index: int, watch: bool):
        replica_id = f"{job_id}-sh-{node_index}-0"
        configuration = client.Configuration()
        configuration.host = self.config["host"]
        configuration.api_key["authorization"] = self.config["bearer_token"]
        configuration.api_key_prefix["authorization"] = "Bearer"
        configuration.verify_ssl = False
        configuration.assert_hostname = False
        client.Configuration.set_default(configuration)
        core_v1 = client.CoreV1Api()

        if watch:
            w = k8swatch.Watch()
            for e in w.stream(
                core_v1.read_namespaced_pod_log,
                name=replica_id,
                namespace="default",
            ):
                yield e
        else:
            api_response = core_v1.read_namespaced_pod_log(
                name=replica_id, namespace="default", follow=watch
            )
            yield api_response

    def shell(self, job_id: str, node_index: int):
        replica_id = f"{job_id}-sh-{node_index}-0"
        subprocess.call(
            [
                "kubectl",
                "--kubeconfig",
                "/dev/null",
                "--insecure-skip-tls-verify=true",
                "--server",
                self.config["host"],
                "--token",
                self.config["bearer_token"],
                "exec",
                "--stdin",
                "--tty",
                replica_id,
                "--",
                "/bin/sh",
            ]
        )

    def cancel(self, job_id: str):
        pass

    def delete(self, job_id: str):
        pass

    @staticmethod
    def supports_shell() -> bool:
        return True

    def prepare_image(
        self,
        base_image: str,
        project_dir: str,
        workspace: Optional[str],
        workdir: Optional[str],
    ) -> str:
        builder = ImageBuilder(base_image, project_dir)

        if workspace:
            if not workdir:
                raise ClientException(
                    "Workspace was provided but no workdir was provided"
                )
            builder.add_overlay(workspace, workdir)

        if workdir:
            builder.set_workdir(workdir)

        image = builder.build_image(self.config["docker_repo"])
        builder.push_image(image, self.config["docker_repo"])
        return image.tags[0]



import click

import komodo_cli.printing as printing
from komodo_cli.utils import APIClient, handle_errors


@click.command("delete-resource")
@click.option(
    "--backend",
    "-b",
    type=str,
    default=None,
    help="Name of the backend containing the resource to delete",
)
@click.option(
    "--resource",
    "-r",
    type=str,
    default=None,
    help="Name of the resource to delete",
)
@click.pass_context
@handle_errors
def cmd_delete_resource(
    ctx: click.Context,
    backend: str,
    resource: str,
):
    api_client: APIClient = ctx.obj["api_client"]

    printing.warning(
        "Deleting a resource will delete all jobs and machines associated with this resource. Are you sure you want to continue?",
        nl=False,
    )
    confirmed = click.confirm("")
    if not confirmed:
        return

    api_client.delete_backend_resource(
        backend,
        resource,
    )

    printing.success(
        f"Resource '{resource}' succesfully deleted from backend '{backend}'"
    )



import os

import click

import komodo_cli.printing as printing
from komodo_cli.utils import APIClient, handle_errors


@click.command("create")
@click.pass_context
@handle_errors
def cmd_create(ctx: click.Context):
    """Log into Komodo and create an API key."""
    api_key_file = os.path.join(os.path.expanduser("~"), ".komo", "api-key")
    os.makedirs(os.path.dirname(api_key_file), exist_ok=True)
    if os.path.exists(api_key_file):
        # Prompt the user for action
        if not click.confirm(
            f"The file {api_key_file} already exists. Do you want to replace it?",
            default=False,
        ):
            printing.warning("Operation cancelled.")
            return

    printing.warning(
        f"Enter your Komodo credentials to get the API key",
    )
    email = None
    while email is None:
        email = click.prompt("email", default=None, type=str)

    password = None
    while password is None:
        password = click.prompt(
            "password",
            default=None,
            type=str,
            hide_input=True,
        )

    printing.info(
        f"Fetching Komodo API key...",
    )
    client = APIClient(username=email, password=password)

    api_key_file = os.path.join(os.path.expanduser("~"), ".komo", "api-key")
    os.makedirs(os.path.dirname(api_key_file), exist_ok=True)
    with open(api_key_file, "w") as f:
        f.write(client.access_token)

    printing.header(
        f"Your API key is below. It has also been written to ~/.komo/api-key",
    )
    printing.info(client.access_token)



import json
import os
import subprocess
import sys
import uuid
from collections import defaultdict
from typing import Dict, List, Tuple

import boto3
import click
from loguru import logger

import komodo_cli.printing as printing

# disabling eks for now
SUPPORTED_COMPUTE_SERVICES = ["aws"]  # ["aws", "eks"]
KOMODO_TERRAFORM_STATE_S3_BUCKET_PREFIX = "komodo-terraform-state"
KOMODO_TERRAFORM_LOCK_DYNAMODB_TABLE = "komodo-terraform-state-lock"


def run_cmd(cmd, print_stdout=True) -> Tuple[int, str, str]:
    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
        shell=True,
    )

    if print_stdout:
        for line in iter(proc.stdout.readline, ""):
            click.echo(line)

    stdout, stderr = proc.communicate()
    exit_code = proc.returncode

    return exit_code, stdout, stderr


def get_compute_dir() -> str:
    komo_dir = os.environ.get(
        "KOMODO_DIR",
        os.path.join(os.path.expanduser("~"), ".komo"),
    )

    compute_dir = os.path.join(komo_dir, "compute")
    return compute_dir


def get_vars_filepath() -> str:
    return os.path.join(get_compute_dir(), "komodo.tfvars")


def print_clusters():
    cmd = [
        "terraform",
        f"-chdir={get_compute_dir()}",
        "output",
        "-json",
        "|",
        "jq",
    ]
    exit_code, stdout, stderr = run_cmd(" ".join(cmd), print_stdout=False)
    if exit_code != 0:
        logger.error(stderr)
        printing.error("Error listing compute", bold=True)
        sys.exit(1)

    outputs = json.loads(stdout.strip())
    clusters = defaultdict(dict)
    for k, v in outputs.items():
        s = k.split("_")
        cluster_type = s[0]
        var = ""
        if cluster_type == "eks":
            var = "_".join(s[2:])
        elif cluster_type == "awsbatch":
            var = "_".join(s[1:])
        else:
            cluster_type = "info"
            var = "_".join(s)

        clusters[cluster_type][var] = v

    for cluster_type, cluster_values in clusters.items():
        if "id" in cluster_values and cluster_values["id"]["value"] != "":
            click.echo(
                f"{click.style(cluster_type, bold=True)} - {click.style(cluster_values['id']['value'], fg='green', bold=True)}"
            )
            cluster_values.pop("id")

        for k, v in cluster_values.items():
            if not v["value"]:
                continue
            click.echo(f"{click.style(k, fg='yellow')}: {v['value']}")


def dict_to_txt_file(d: Dict, output_path: str):
    with open(output_path, "w") as file:
        for k, v in d.items():
            if isinstance(v, list) or isinstance(v, dict):
                v = json.dumps(v).replace(":", "=")
            elif isinstance(v, bool):
                v = str(v).lower()
            else:
                if not v.startswith('"') and not v.endswith('"'):
                    v = f'"{v}"'

            file.write(f"{k} = {v}\n")


def txt_to_dict(input_path: str) -> Dict:
    output = {}
    with open(input_path, "r") as file:
        for line in file:
            line = line.strip()
            if line:
                k, v = line.split("=", 1)
                k = k.strip()
                v = v.strip()
                v = v.replace("=", ":")

                if v.startswith("[") and v.endswith("]"):
                    v = json.loads(v)
                if v.startswith("{") and v.endswith("}"):
                    v = json.loads(v)

                output[k] = v
    return output


def get_vars(no_exist_ok=False):
    compute_dir = get_compute_dir()
    vars_path = os.path.join(compute_dir, "komodo.tfvars")

    if not os.path.exists(vars_path):
        if no_exist_ok:
            return {}
        raise FileNotFoundError(f"Variables file not found at {vars_path}")
    else:
        return txt_to_dict(vars_path)


def update_vars(new_vars: Dict):
    compute_dir = get_compute_dir()
    vars_path = os.path.join(compute_dir, "komodo.tfvars")
    dict_to_txt_file(new_vars, vars_path)


def bucket_exists_with_prefix(s3_client, prefix):
    response = s3_client.list_buckets()
    for bucket in response["Buckets"]:
        if bucket["Name"].startswith(prefix):
            return True, bucket["Name"]
    return False, ""


def ensure_terraform_state_resources(aws_profile: str):
    session = boto3.Session(profile_name=aws_profile, region_name="us-east-2")
    s3_client = session.client("s3")
    retval_bucket_name = ""
    retval_dynamodb_table_name = KOMODO_TERRAFORM_LOCK_DYNAMODB_TABLE
    # check if komodo-terraform-state bucket exists, if not create
    komodo_terraform_state_bucket_exists, bucket_name = bucket_exists_with_prefix(
        s3_client=s3_client, prefix=KOMODO_TERRAFORM_STATE_S3_BUCKET_PREFIX
    )
    if komodo_terraform_state_bucket_exists:
        retval_bucket_name = bucket_name
        printing.success(
            f"Terraform state bucket {bucket_name} found.",
            bold=True,
        )
    else:
        printing.warning(
            "Terraform state bucket not found. Creating terraform state bucket."
        )
        unique_suffix = str(uuid.uuid4())
        new_bucket_name = f"{KOMODO_TERRAFORM_STATE_S3_BUCKET_PREFIX}-{unique_suffix}"
        try:
            s3_client.create_bucket(
                Bucket=new_bucket_name,
                CreateBucketConfiguration={"LocationConstraint": "us-east-2"},
            )
            s3_client.put_bucket_versioning(
                Bucket=new_bucket_name,
                VersioningConfiguration={"Status": "Enabled"},
            )
            printing.success(
                f"Terraform state bucket {new_bucket_name} created successfully.",
                bold=True,
            )
        except Exception as e:
            logger.error(e)
            printing.error(
                f"Error creating terraform state bucket.",
                bold=True,
            )
            sys.exit(1)

        retval_bucket_name = new_bucket_name

    dynamodb_client = session.client("dynamodb")
    # check if komodo-terraform-state-lock table exists, if not create
    try:
        dynamodb_client.describe_table(TableName=KOMODO_TERRAFORM_LOCK_DYNAMODB_TABLE)
        printing.success(
            f"Terraform state locking table {KOMODO_TERRAFORM_LOCK_DYNAMODB_TABLE} found.",
            bold=True,
        )
    except Exception as e:
        printing.warning(
            "Terraform state locking table not found. Creating terraform state locking table."
        )
        try:
            dynamodb_client.create_table(
                TableName=KOMODO_TERRAFORM_LOCK_DYNAMODB_TABLE,
                KeySchema=[
                    {"AttributeName": "LockID", "KeyType": "HASH"},
                ],
                AttributeDefinitions=[
                    {"AttributeName": "LockID", "AttributeType": "S"},
                ],
                ProvisionedThroughput={
                    "ReadCapacityUnits": 20,
                    "WriteCapacityUnits": 20,
                },
            )
            printing.success(
                f"Terraform state locking table ${KOMODO_TERRAFORM_LOCK_DYNAMODB_TABLE} created successfully.",
                bold=True,
            )
        except Exception as e:
            logger.error(e)
            printing.error(
                f"Error creating terraform state locking table.",
                bold=True,
            )
            sys.exit(1)

    return retval_bucket_name, retval_dynamodb_table_name



import textwrap

import click
import yaml
from tabulate import tabulate

import komodo_cli.printing as printing
from komodo_cli.utils import APIClient, handle_errors


@click.command("list")
@click.option(
    "--configs", "-c", is_flag=True, help="Display the config for each backend"
)
@click.pass_context
@handle_errors
def cmd_list(
    ctx: click.Context,
    configs: bool,
):
    api_client: APIClient = ctx.obj["api_client"]

    backends = api_client.list_backends()

    backends_to_print = []
    for backend in backends:
        if len(backend.config) == 0:
            config_str = ""
        else:
            config_strs = []
            for key, value in backend.config.items():
                config_strs.append(f"{key}: {value}")

            config_str = "\n".join(config_strs)
            config_str = "\n".join(
                textwrap.wrap(
                    config_str,
                    80,
                    break_on_hyphens=False,
                    replace_whitespace=False,
                    drop_whitespace=False,
                )
            )

        if len(backend.resources) == 0:
            resources = ""
        else:
            resources = "\n".join(
                [yaml.dump({r.name: r.config}) for r in backend.resources]
            )
        print_info = [backend.name, backend.type, config_str, resources]
        backends_to_print.append(print_info)

    if configs:
        headers = ["Name", "Type", "Config", "Resources"]
        maxcolwidths = [None, None, None, None]
    else:
        headers = ["Name", "Type", "Resources"]
        maxcolwidths = [None, None, None]
        backends_to_print = [row[0:2] + row[3:] for row in backends_to_print]

    if len(backends_to_print) == 0:
        maxcolwidths = None  # weird tabulate bug

    printing.header("Backends:")
    # TODO: use a better library than tabulate to make this pretty
    printing.info(
        tabulate(
            backends_to_print,
            headers=headers,
            maxcolwidths=maxcolwidths,
            tablefmt="simple_grid",
        ),
    )



import json
import os
import pathlib
import shutil
import sys
from collections import defaultdict

import click
from loguru import logger
from tabulate import tabulate

import komodo_cli.printing as printing
from komodo_cli.commands.compute.utils import get_compute_dir, print_clusters
from komodo_cli.utils import handle_errors


@click.command("list")
@click.pass_context
@handle_errors
def cmd_list(ctx: click.Context):
    """List Komodo computes"""
    compute_dir = get_compute_dir()

    compute_dir = get_compute_dir()
    script_dir = f"{pathlib.Path(__file__).parent.resolve()}"
    if not os.path.exists(compute_dir):
        os.makedirs(compute_dir)
    tf_dir = os.path.abspath(os.path.join(script_dir, "..", "..", "terraform"))
    shutil.copytree(tf_dir, compute_dir, dirs_exist_ok=True)

    if not os.path.exists(compute_dir):
        printing.error(f"No compute configuration found.", bold=True)
        sys.exit(1)
    else:
        print_clusters()



import os
import sys

import click
import git
import jinja2

import komodo_cli.printing as printing
from komodo_cli.utils import APIClient

BACKEND_CONFIG_TEMPLATE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "../..", "templates"
)


@click.command("project")
@click.pass_context  # TODO: do we need ctx?
# what happens if the cli.py fails to read these configs?
def cmd_project(ctx: click.Context):
    """Initialize a Komodo project in the current directory."""
    # TODO: if git repo doesn't exist, use cwd
    git_repo = git.Repo(os.getcwd(), search_parent_directories=True)
    git_root = git_repo.git.rev_parse("--show-toplevel")
    if os.path.isfile(os.path.join(git_root, ".komo", "project.yaml")):
        printing.error(
            f"Project config already exists in this repository",
            bold=True,
        )
        sys.exit(1)

    printing.header(
        f"Initializing Komo project config in {git_root}/.komo ",
        bold=True,
    )

    api_client: APIClient = ctx.obj["api_client"]
    backend_schemas = api_client.list_backends()

    template_loader = jinja2.FileSystemLoader(
        searchpath=os.path.join(BACKEND_CONFIG_TEMPLATE_DIR)
    )
    template_env = jinja2.Environment(loader=template_loader)
    project_template = template_env.get_template("project.jinja")

    backend_templates = {}
    for backend_schema in backend_schemas:
        backend_template = template_env.get_template(
            f"{backend_schema.type}.jinja",
        )
        backend_templates[backend_schema.name] = backend_template.render()

    project_config_output = project_template.render(
        backends=backend_templates,
        first_backend=backend_schemas[0].name,
    )

    project_config_filepath = os.path.join(git_root, ".komo", "project.yaml")
    os.makedirs(os.path.dirname(project_config_filepath), exist_ok=True)
    with open(project_config_filepath, "w") as f:
        f.write(project_config_output)

    printing.success(
        f"Komodo project successfully initialized at project root.\nUpdate the .komo/project.yaml file to your needs.",
        bold=True,
    )



import os
import pathlib
import shutil
import sys
from typing import Optional

import click
import jinja2
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.commands.compute.utils import (
    SUPPORTED_COMPUTE_SERVICES, ensure_terraform_state_resources,
    get_compute_dir, get_vars, get_vars_filepath, print_clusters, run_cmd,
    update_vars)
from komodo_cli.utils import handle_errors


@click.command("create")
@click.option(
    "--compute-type",
    "-t",
    type=str,
    default="aws",
    help="The type of compute infrastructure type to create.",
)
@click.option(
    "--name",
    "-n",
    type=str,
    required=True,
    help="The name for the newly created compute infrastructure",
)
@click.option(
    "--aws-region",
    type=str,
    required=True,
    help="The AWS region to create the compute infrastructure in.",
)
@click.option(
    "--aws-profile",
    type=str,
    required=True,
    help="The AWS profile to use for creating the compute infrastructure.",
)
@click.option(
    "--manual-confirm",
    "-m",
    type=bool,
    is_flag=True,
    default=False,
    required=False,
    help="Manually confirm terraform plan.",
)
@click.pass_context
@handle_errors
def cmd_create(
    ctx: click.Context,
    compute_type: Optional[str],
    name: Optional[str],
    aws_region: Optional[str],
    aws_profile: Optional[str],
    manual_confirm: Optional[bool],
):
    """Create Komodo compute"""
    if compute_type not in SUPPORTED_COMPUTE_SERVICES:
        printing.error(
            f"Unsupported compute type {compute_type}. Supported compute types: {SUPPORTED_COMPUTE_SERVICES}",
            bold=True,
        )
        sys.exit(1)

    compute_dir = get_compute_dir()
    script_dir = f"{pathlib.Path(__file__).parent.resolve()}"
    if not os.path.exists(compute_dir):
        os.makedirs(compute_dir)
    tf_dir = os.path.abspath(os.path.join(script_dir, "..", "..", "terraform"))

    (
        terraform_state_bucket_name,
        terraform_state_lock_dynamodb_table_name,
    ) = ensure_terraform_state_resources(aws_profile)

    # generate providers.tf from template
    templateLoader = jinja2.FileSystemLoader(searchpath=tf_dir)
    templateEnv = jinja2.Environment(loader=templateLoader)
    TEMPLATE_FILE = "providers.jinja"
    template = templateEnv.get_template(TEMPLATE_FILE)
    outputText = template.render(
        aws_profile=aws_profile,
        terraform_state_bucket_name=terraform_state_bucket_name,
        terraform_state_lock_dynamodb_table_name=terraform_state_lock_dynamodb_table_name,
    )
    with open(os.path.join(tf_dir, "providers.tf"), "w") as f:
        f.write(outputText)

    shutil.copytree(
        tf_dir,
        compute_dir,
        dirs_exist_ok=True,
        ignore=shutil.ignore_patterns("*.jinja"),
    )

    cmd = ["terraform", f"-chdir={compute_dir}", "init"]
    exit_code, _, stderr = run_cmd(" ".join(cmd))
    if exit_code != 0:
        logger.error(stderr)
        printing.error(f"Error initializing terraform.", bold=True)
        exit(1)

    try:
        tf_vars = get_vars(no_exist_ok=True)
    except Exception as e:
        logger.error(e)
        printing.error(
            f"Error loading tf vars",
            bold=True,
        )
        sys.exit(1)

    tf_vars["aws_region"] = aws_region
    tf_vars["aws_profile"] = aws_profile

    if compute_type == "eks":
        tf_vars[f"{compute_type}_cluster_id"] = name
    elif compute_type == "aws":
        tf_vars["create_aws_batch"] = True
        tf_vars["aws_batch_name"] = name

    update_vars(tf_vars)

    cmd = [
        "terraform",
        f"-chdir={compute_dir}",
        "apply",
    ]
    if not manual_confirm:
        cmd.append("-auto-approve")
    cmd.append(f"-var-file={get_vars_filepath()}")

    exit_code, _, stderr = run_cmd(" ".join(cmd))

    if exit_code != 0:
        logger.error(stderr)
        printing.error("Error creating compute", bold=True)
        printing.error(stderr)
        sys.exit(1)

    printing.success("Compute created successfully", bold=True)
    print_clusters()
    # TODO: post backend details to the server for backend config



import os
import uuid
from typing import Optional

import docker
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.types import ImageBuildException, ImagePushException

KOMODO_DOCKERFILE_NAME = "komodo.Dockerfile"


class ImageBuilder:
    def __init__(self, base_image, project_dir):
        self.dockerfile = f"FROM {base_image}"
        requirements_file = os.path.join(project_dir, "requirements.txt")
        if os.path.isfile(requirements_file):
            self.dockerfile += "\nCOPY requirements.txt /tmp/requirements.txt"
            if base_image.startswith("nvcr.io/nvidia"):
                # uninstall opencv to allow the user to install it without issues
                self.dockerfile += "\nRUN pip uninstall -y opencv"
                self.dockerfile += (
                    "\nRUN rm -rf /usr/local/lib/python*/dist-packages/cv2/"
                )
                self.dockerfile += "\nRUN apt update -y"
                self.dockerfile += "\nRUN apt install -y libgl1-mesa-glx"
            self.dockerfile += (
                "\nRUN pip install --no-cache-dir -r /tmp/requirements.txt"
            )
        self.project_dir = project_dir
        self.client = docker.from_env()

    def add_overlay(self, overlay_dir, dest_dir):
        self.dockerfile += f"\nCOPY {overlay_dir} {dest_dir}"

    def set_workdir(self, workdir):
        self.dockerfile += f"\nWORKDIR {workdir}"

    def add_aws_cli(self):
        self.dockerfile += '\nRUN command -v aws >/dev/null 2>&1 || { curl "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip" -o "awscliv2.zip"; unzip awscliv2.zip; ./aws/install --bin-dir /usr/local/bin --install-dir /usr/local/aws-cli --update; rm -rf aws; rm -rf awscliv2.zip; }'

    def add_aws_efa(self):
        self.dockerfile += "\nRUN apt update"
        self.dockerfile += "\nRUN curl -O https://efa-installer.amazonaws.com/aws-efa-installer-1.30.0.tar.gz"
        self.dockerfile += "\nRUN tar -xf aws-efa-installer-1.30.0.tar.gz && cd aws-efa-installer && ./efa_installer.sh -y --mpi=openmpi4 --skip-kmod --skip-limit-conf --no-verify"
        self.dockerfile += "\nRUN apt-get install -y libhwloc-dev"
        self.dockerfile += "\nRUN wget https://github.com/aws/aws-ofi-nccl/releases/download/v1.7.4-aws/aws-ofi-nccl-1.7.4-aws.tar.gz && tar -xf aws-ofi-nccl-1.7.4-aws.tar.gz && cd aws-ofi-nccl-1.7.4-aws && ./configure --prefix=/opt/aws-ofi-nccl --with-mpi=/opt/amazon/openmpi --with-libfabric=/opt/amazon/efa --with-cuda=/usr/local/cuda --enable-platform-aws"
        self.dockerfile += '\nENV PATH="/opt/amazon/openmpi/bin/:${PATH}"'
        self.dockerfile += "\nRUN make && make install; exit 0"

    def build_image(self, overlay_images_repository: Optional[str] = None):
        printing.info("Building image...")
        dockerfile_path = os.path.join(self.project_dir, KOMODO_DOCKERFILE_NAME)
        with open(dockerfile_path, "w") as f:
            f.write(self.dockerfile)

        uid = str(uuid.uuid4())
        if overlay_images_repository:
            tag = f"{overlay_images_repository}:{uid}"
        else:
            tag = f"komodo-overlay:{uid}"

        try:
            image, build_logs = self.client.images.build(
                path=os.path.dirname(dockerfile_path),
                tag=tag,
                quiet=False,
                rm=True,
                pull=True,
                forcerm=True,
                dockerfile=dockerfile_path,
                platform="linux/amd64",  # TODO
            )
            return image
        except (
            docker.errors.BuildError,
            docker.errors.APIError,
            TypeError,
        ) as e:
            raise ImageBuildException(str(e))
        except Exception as e:
            raise e
        finally:
            os.remove(dockerfile_path)

    def push_image(self, image, overlay_images_repository):
        printing.info("Pushing image...")
        try:
            resp = self.client.images.push(
                overlay_images_repository,
                image.tags[0].split(":")[-1],
                stream=True,
                decode=True,
            )

            for line in resp:
                if "error" in line:
                    raise ImagePushException(line["error"])
        except ImagePushException as e:
            raise e
        except Exception as e:
            raise ImagePushException(str(e))
